import collections
import json

import padl
import torch


@padl.transform
class Trainer:
    def __init__(self, train_model, optimizer, infer_model=None,
                 metrics=None, iterate_args=None):
        self.train_model = train_model
        self.infer_model = infer_model
        self.optimizer = optimizer
        self.metrics = metrics
        if iterate_args is None:
            self.iterate_args = {}
        self.metric_data = {}
        self.iteration = 0
        self.epoch = 0

    def pre_save(self, path, i):
        print(f'saving optimizer state to {path}/{i}.pt')
        torch.save(self.optimizer.state_dict(), path / f'{i}.pt')

        print(f'saving iterate args to {path}/{i}.iterate.json')
        with open(path / f'{i}.iterate.json', 'w') as f:
            json.dump(self.iterate_args, f)

        print(f'saving metrics to {path}/{i}.metrics.json')
        with open(path / f'{i}.metrics.json', 'w') as f:
            json.dump(self.metric_data, f)

        print(f'saving iteration to {path}/{i}.iteration.json')
        with open(path / f'{i}.iteration.json', 'w') as f:
            json.dump(self.iteration, f)

        print(f'saving epoch to {path}/{i}.epoch.json')
        with open(path / f'{i}.epoch.json', 'w') as f:
            json.dump(self.epoch, f)

    def post_load(self, path, i):
        sd = torch.load(path / f'{i}.pt', map_location=self.pd_device)
        print('loading optimizer from state-dict')
        self.optimizer.load_state_dict(sd)

        print(f'loading iterate args')
        with open(path / f'{i}.iterate.json') as f:
            self.iterate_args.update(json.load(f))

        print(f'loading metrics')
        with open(path / f'{i}.metrics.json') as f:
            self.metric_data = json.load(f)

        print(f'loading iteration')
        with open(path / f'{i}.iteration.json') as f:
            self.iteration = json.load(f)
        self.iteration += 1

        print(f'loading epoch')
        with open(path / f'{i}.epoch.json') as f:
            self.epoch = json.load(f)
        self.epoch += 1

    def __call__(self, *args, **kwargs):
        raise NotImplementedError

    def train_step(self, output):
        self.optimizer.zero_grad()
        output.backward()
        self.optimizer.step()

    def aggregate_losses(self, losses):
        return sum([x.item() for x in losses]) / len(losses)

    def log_train(self, output):
        print(f'TRAIN epoch {self.epoch}; iteration {self.iteration}; loss {output};')

    def log_valid(self, metrics):
        str_ = ''.join([f'{k}: {v[-1]};' for k, v in metrics.items()])
        print(f'TRAIN epoch {self.epoch}; iteration {self.iteration};' + str_)

    def save(self, path, *args, **kwargs):
        self.pd_save(path, force_overwrite=True)

    def watch_function(self, metrics):
        try:
            return metrics['loss'][-1] == min(metrics['loss'])
        except (KeyError, IndexError):
            return True

    def train(self, train_data, path, valid_data=None, metric_data=None, ground_truth=None,
              n_epochs=100, max_iterations=1000000, save_interval=None,
              valid_interval=100, **kwargs):

        self.iterate_args.update(kwargs)

        metrics = collections.defaultdict(lambda: [])
        metrics.update(self.metric_data)

        for epoch in range(self.epoch, n_epochs):
            for output in self.train_model.train_apply(train_data, **self.iterate_args):
                self.train_step(output)
                self.log_train(output)

                if valid_data is None:
                    assert save_interval is not None
                    if self.iteration % save_interval == 0:
                        self.save(path, self.iteration, epoch)
                    self.iteration += 1
                    continue

                if self.iteration and self.iteration % valid_interval == 0 and valid_data is not None:
                    valid_loss = []
                    for output in self.train_model.eval_apply(valid_data, **self.iterate_args):
                        valid_loss.append(output)
                    valid_loss = self.aggregate_losses(valid_loss)
                    metrics['loss'].append(valid_loss)

                    if self.infer_model is not None and self.metrics is not None:
                        assert len(self.metrics)
                        predictions = list(self.infer_model.eval_apply(metric_data, **self.iterate_args))
                        for metric in self.metrics:
                            metrics[metric].append(self.metrics[metric](predictions, ground_truth))

                    self.log_valid(metrics)
                    if self.watch_function(metrics):
                        self.metric_data = metrics
                        self.save(path, self.iteration, epoch)

                if self.iteration >= max_iterations:
                    return
                self.iteration += 1




import os
import subprocess
import pathlib


def prepare(target_model, version='1.0', force=False):
    """Prepare and package model into archive `mar` file with TorchModelArchiver.

    :param target_model: PADL serialized model - `.padl`
    :param version: version of model
    :param force: if force=True, existing archive file is overwritten
    """

    target_model = pathlib.Path(target_model)
    model_parent = target_model.parents[0]
    model_name = target_model.stem
    current_directory = pathlib.Path(__file__).parent
    handler_dir = current_directory/'handler.py'

    print(f'handler is {handler_dir}')
    print(f'converting {model_name} to MAR format...')

    cmd = [
        'torch-model-archiver',
        '--model-name', model_name,
        '--version', version,
        '--export-path', model_parent,
        '--extra-files', target_model,
        '--handler', handler_dir,
    ]
    if force:
        cmd += ['--force']
    subprocess.run(cmd)


def serve(model_store,
          model,
          ncs=True,
          workflow_store=None,
          ts_config=None,
          log_config=None):
    """Serve model with TorchServe.

    :param model_store: directory where model is stored
    :param model: model file name - `.mar`
    :param ncs: Disable snapshot feature
    :param ts_config: Configuration file for TorchServe
    :param workflow_store: Workflow store location where workflow can be loaded. Defaults to model_store
    :param log_config: Log4j configuration file for TorchServe
    """
    cmd = [
        'torchserve',
        '--start',
        '--model-store', model_store,
        '--models', model,
    ]
    if ncs:
        cmd += ['--ncs']
    if workflow_store is not None:
        cmd += ['--workflow-store', workflow_store]
    if ts_config is not None:
        cmd += ['--ts-config', ts_config]
    if log_config is not None:
        cmd += ['--log-config', log_config]
    cmd += ['--foreground']
    subprocess.run(cmd)


def stop():
    """Stop TorchServe model-server."""
    subprocess.run(['torchserve', '--stop'])


def prepare_and_serve(target_model,
                      version='1.0',
                      force=False,
                      ncs=True,
                      workflow_store=None,
                      ts_config=None,
                      log_config=None):
    """Package model and serve with TorchServe.

    :param target_model: PADL serialized model - `.padl`
    :param version: version of model
    :param force: if force=True, existing archive file is overwritten
    :param ncs: Disable snapshot feature for TorchServe
    :param ts_config: Configuration file for TorchServe
    :param workflow_store: Workflow store location for TorchServe where workflow can be loaded. Defaults to model_store
    :param log_config: Log4j configuration file for TorchServe
    """
    target_model = pathlib.Path(target_model)
    model_parent = target_model.parent
    model_name = target_model.stem
    mar_name = pathlib.Path(f'{model_name}.mar')

    if force or (not os.path.exists(model_parent/mar_name)):
        prepare(target_model, version=version, force=force)
    else:
        print(f'Torch model archive already exists for {target_model}, skipping archiving...')

    print(f'converting {model_name} to MAR format...')
    print('Serving the model...')

    serve(model_parent, mar_name, ncs=ncs, workflow_store=workflow_store, ts_config=ts_config, log_config=log_config)



try:
    from padl_ext.pytorch_lightning.prepare import LightningModule
except ImportError:
    print("Please install pytorch lightning dependencies "
          "(pip install padl-extensions[pytorch_lightning]) to use the connector")
    pass

try:
    from padl_ext.torchserve.prepare import prepare
    from padl_ext.torchserve.handler import PadlHandler
except ImportError:
    print("Please install torchserve dependencies "
          "(pip install padl-extensions[torchserve]) to use the connector")
    pass




import pytest
import torch
import tempfile
import padl_ext.torchserve as torchserve
import padl

from tests.material import utils


@pytest.fixture()
def tmp_dir():
    tmp_path = tempfile.TemporaryDirectory(dir='tests/material/')
    yield tmp_path.name
    tmp_path.cleanup()


@padl.transform
class PadlModel(torch.nn.Module):
    def __init__(self):
        super().__init__()
        self.encoder = torch.nn.Sequential(
            torch.nn.Linear(36, 128), torch.nn.ReLU(), torch.nn.Linear(128, 3)
        )
        self.decoder = torch.nn.Sequential(
            torch.nn.Linear(3, 3), torch.nn.Linear(3, 10)
        )

    def forward(self, x):
        x = self.encoder(x)
        x = self.decoder(x)
        return x


@padl.transform
def postprocess(x):
    return torch.argmax(torch.nn.Softmax(dim=1)(x), dim=1)


@pytest.mark.skipif((not utils.check_if_module_installed('torchserve')) and \
                    (not utils.check_if_module_installed("torch-model-archiver")),
                    reason="requires the torchserve and torch-model-archiver")
class TestTorchServe:
    @pytest.fixture(autouse=True, scope='class')
    def init(self, request):
        request.cls.transform_1 = PadlModel() >> postprocess

    def test_torchserve(self, tmp_dir):
        save_dir = tmp_dir + '/temp.padl'
        padl.save(self.transform_1, save_dir)
        torchserve.prepare_and_serve(save_dir)
        torchserve.stop()



"""Connector to Pytorch Lightning"""

import shutil
from pathlib import Path
import torch
import padl

import pytorch_lightning as pl


def _padl_data_loader(data, padl_model, mode, **kwargs):
    """Create the :class:`~torch.utils.data.DataLoader` used by PADL models.

    This creates a :class:`~torch.utils.data.DataLoader` that can be directly passed to the
    :meth:`pytorch_lightning.Trainer.fit` function.

    Example:

    >>> from padl import transform, batch, identity
    >>> import torch
    >>> from torch.utils.data import DataLoader
    >>> import pytorch_lightning as pl
    >>> @transform
    ... class Net(torch.nn.Module):
    ...     def __init__(self):
    ...         super().__init__()
    ...         self.encoder = torch.nn.Linear(16, 16)
    ...     def forward(self, x):
    ...         return self.encoder(x)
    >>> @transform
    ... def padl_loss(reconstruction, original):
    ...     return torch.nn.functional.mse_loss(reconstruction, original)
    >>> model = identity >> batch >> Net() + identity >> padl_loss
    >>> data_list = [torch.randn([16])] * 4
    >>> train_data_loader = _padl_data_loader(data_list, model, 'train', batch_size=2)
    >>> isinstance(train_data_loader, DataLoader)
    True
    >>> val_data_loader = _padl_data_loader(data_list, model, 'eval', batch_size=2)
    >>> isinstance(val_data_loader, DataLoader)
    True
    >>> padl_lightning = LightningModule(model)
    >>> trainer = pl.Trainer()
    >>> trainer.fit(padl_lightning, train_data_loader, val_data_loader)

    :param data: List or iterator of data points to be preprocessed by `padl_model`
    :param padl_model: PADL transform to be used in training
    :param mode: PADL mode to call the preprocess :class:`padl.transforms.Transform` in
        (can be 'infer', 'eval' or 'train')
    :param kwargs: Keyword arguments passed to the data loader
        (see :class:`~torch.utils.data.DataLoader` documentation for details).
    """
    return padl_model.pd_get_loader(data, padl_model.pd_preprocess, mode, **kwargs)


@padl.transform
class LightningModule(pl.LightningModule):
    """PADL connector to Pytorch Lightning.

    :param padl_model: PADL transform. Can provide a string path to load a PADL transform.
    :param train_data: list of training data points
    :param val_data: list of validation data points
    :param test_data: list of test data points
    :param learning_rate: learning rate
    :param kwargs: loader key word arguments for the DataLoader
    """
    pd_save_options = {torch.nn.Module: 'no-save'}
    def __init__(
        self,
        padl_model,
        trainer,
        train_data=None,
        val_data=None,
        test_data=None,
        learning_rate=1e-4,
        inference_model=None,
        **kwargs
    ):
        super().__init__()

        if isinstance(padl_model, str):
            padl_model = padl.load(padl_model)
        elif not isinstance(padl_model, padl.transforms.Transform):
            raise TypeError('Please provide a PADL transform or a str path to load a '
                            'PADL transform')

        self.padl_model = padl_model

        self.train_data = train_data
        self.val_data = val_data
        self.test_data = test_data
        self.loader_kwargs = kwargs
        self.learning_rate = learning_rate
        self.pd_previous = []
        self.trainer = trainer

        self.padl_model.pd_forward_device_check()

        # Set Pytorch layers as attributes from PADL model
        layers = self.padl_model.pd_layers
        for layer in layers:
            if layer.pd_name is not None:
                prefix = f'{layer.pd_name}'
            else:
                prefix = f'{layer.__class__.__name__}'
            key = prefix
            counter = 0
            while hasattr(self, key):
                key = f'{prefix}_{counter}'
                counter += 1
            setattr(self, key, layer)

        self.inference_model = inference_model

    def fit(self, *args, train_data=None, val_data=None, **kwargs):
        if train_data is not None:
            self.train_data = train_data
        if val_data is not None:
            self.val_data = val_data
        if hasattr(self, '_restore_path'):
            assert 'ckpt_path' not in kwargs, 'ckpt_path not supported'
            self.trainer.fit(self, *args, **kwargs, ckpt_path=self._restore_path)
        else:
            self.trainer.fit(self, *args, **kwargs)

    def forward(self, x):
        """In pytorch lightning, forward defines the prediction/inference actions."""
        return None

    def train_dataloader(self):
        """Creates the train :class:`~torch.utils.data.DataLoader`
        using :meth:`self.padl_model.pd_get_loader` if `self.train_data` is provided.
        """
        if self.train_data is not None:
            return self.padl_model.pd_get_loader(self.train_data, self.padl_model.pd_preprocess,
                                                 'train', **self.loader_kwargs)
        return None

    def val_dataloader(self):
        """Creates the validation :class:`~torch.utils.data.DataLoader`
        using :meth:`self.padl_model.pd_get_loader` if `self.val_data` is provided.
        """
        if self.val_data is not None:
            return self.padl_model.pd_get_loader(self.val_data, self.padl_model.pd_preprocess,
                                                 'eval', **self.loader_kwargs)
        return None

    def test_dataloader(self):
        """Creates the test :class:`~torch.utils.data.DataLoader`
        using :meth:`self.padl_model.pd_get_loader` if `self.test_data` is provided.
        """
        if self.test_data is not None:
            return self.padl_model.pd_get_loader(self.test_data, self.padl_model.pd_preprocess,
                                                 'eval', **self.loader_kwargs)
        return None

    def training_step(self, batch, batch_idx):
        """Default training step overwritten from
        :meth:`pytorch_lightning.LightningModule.training_step`.

        Note: The data loader generated by :meth:`self.padl_model.pd_get_loader` returns a
            tuple of (idx, batch). We only need the batch here.

        :param batch: The output of your :class:`~torch.utils.data.DataLoader`.
            A tensor, tuple or list.
        :param batch_idx: Integer displaying index of this batch
        """
        _, batch = batch
        loss = self.padl_model.pd_forward.pd_call_in_mode(batch, 'train')
        self.log("train_loss", loss)
        return loss

    def validation_step(self, batch, batch_idx):
        """Default validation step overwritten from
        :meth:`pytorch_lightning.LightningModule.validation_step`.

        Note: The data loader generated by :meth:`self.padl_model.pd_get_loader` returns a
            tuple of (idx, batch). We only need the batch here.

        :param batch: The output of your :class:`~torch.utils.data.DataLoader`.
            A tensor, tuple or list.
        :param batch_idx: Integer displaying index of this batch
        """
        _, batch = batch
        loss = self.padl_model.pd_forward.pd_call_in_mode(batch, 'eval')
        self.log("val_loss", loss)

    def test_step(self, batch, batch_idx):
        """Default test step overwritten from :meth:`pytorch_lightning.LightningModule.test_step`.

        Note: The data loader generated by :meth:`self.padl_model.pd_get_loader` returns a
            tuple of (idx, batch). We only need the batch here.

        :param batch: The output of your :class:`~torch.utils.data.DataLoader`.
            A tensor, tuple or list.
        :param batch_idx: Integer displaying index of this batch
        """
        _, batch = batch
        loss = self.padl_model.pd_forward.pd_call_in_mode(batch, 'eval')
        self.log("test_loss", loss)

    def on_save_checkpoint(self, checkpoint):
        """Called by Lightning when saving a checkpoint to give you a chance to store anything
        else you might want to save.

        Adding PADL saving by implementing the
        :meth:`pytorch_lightning.LightningModule.on_save_checkpoint` callback and calling
        :meth:`self.padl_model.save`.
        This works in addition to the :class:`pytorch_lightning.callbacks.ModelCheckpoint`
        callback which will still save the pytorch lightning ckpt file.

        :param checkpoint: The full checkpoint dictionary before it gets dumped to a file.
            Implementations of this hook can insert additional data into this dictionary.
        """
        callback_state_keys = checkpoint['callbacks'].keys()
        callback_state_keys = [str(k) for k in callback_state_keys]
        checkpoint_callback_key = [k for k in callback_state_keys if 'ModelCheckpoint' in k][0]
        checkpoint_callback = checkpoint['callbacks'][checkpoint_callback_key]

        best_model_path = checkpoint_callback.get('best_model_path')
        best_model_path = best_model_path.replace(Path(best_model_path).suffix, '')

        best_k_model_paths = checkpoint_callback.get('best_k_models')
        if best_k_model_paths is not None:
            best_k_model_paths = [x.replace(Path(x).suffix, '') for x in best_k_model_paths.keys()]
        else:
            best_k_model_paths = []

        if len(best_k_model_paths) == 0:
            path = best_model_path
        else:
            path = best_k_model_paths[-1]

        if path not in self.pd_previous:
            self.pd_previous.append(path)
            self.pd_save(path, force_overwrite=True)

        if len(best_k_model_paths) == 0:
            k = 1
        else:
            k = len(best_k_model_paths)

        del_dirpath = None
        if len(self.pd_previous) == k + 1:
            del_dirpath = self.pd_previous.pop(0) + '.padl'

        if del_dirpath is not None:
            shutil.rmtree(del_dirpath)

        # This will get saved in the Pytorch Lightning ckpt and is needed for reloading the ckpt
        checkpoint['padl_models'] = self.pd_previous

    @property
    def best_model_path(self):
        return self.trainer.checkpoint_callback.best_model_path.split('.ckpt')[0] + '.padl'

    @property
    def best_k_models(self):
        return [x.split('.ckpt')[0] + '.padl'
                for x in self.trainer.checkpoint_callback.best_k_models]

    def post_load(self, path, i):
        self._restore_path = str(path).split('.padl')[0] + '.ckpt'
        self.load_state_dict(torch.load(self._restore_path)['state_dict'])

    def configure_optimizers(self):
        """Implementation of the inherited method
        :meth:`pytorch_lightning.LightningModule.configure_optimizers`.

        A default optimizer is provided for ease of use, but feel free to overwrite for your
        specific use case. You can choose what optimizers and learning-rate schedulers to use
        in your optimization. Normally you'd need one. But in the case of GANs or similar you
        might have multiple.
        """
        optimizer = torch.optim.Adam(self.parameters(), lr=self.learning_rate)
        return optimizer




import json
import logging

from ts.torch_handler.base_handler import BaseHandler
from padl import load, unbatch
from padl.transforms import Transform

logger = logging.getLogger(__name__)



class PadlHandler(BaseHandler):
    def __init__(self):
        super().__init__()
        self._pd_preprocess = None
        self._pd_forward = None
        self._pd_postprocess = None

    def initialize(self, context):
        properties = context.system_properties
        model_dir = properties.get('model_dir')
        logger.warning(model_dir)
        m = load(model_dir)
        self._pd_preprocess = m.pd_preprocess
        self._pd_forward = m.pd_forward
        self._pd_postprocess = m.pd_postprocess

    def preprocess(self, data):
        data = data[0]['body']
        return self._pd_preprocess.infer_apply(data)

    def inference(self, data, *args, **kwargs):
        return self._pd_forward.infer_apply(data)

    def postprocess(self, data):
        Transform.pd_mode = 'infer'
        output =self._pd_postprocess[1:].infer_apply(unbatch(data))
        return [json.dumps(output)]



import setuptools

import pkg_resources
import pathlib
from distutils.util import convert_path

versions = {}
ver_path = convert_path('padl_ext/version.py')
with open(ver_path) as ver_file:
    exec(ver_file.read(), versions)

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read().split('\n')

long_description = ['# PADL-EXTENSIONS\n'] + [x for x in long_description if not x.strip().startswith('<img') and not x.strip().startswith('[!')]
long_description = '\n'.join(long_description)


def parse_requirements(filename):
    with pathlib.Path(filename).open() as requirements_txt:
        return [str(requirement) for requirement in pkg_resources.parse_requirements(requirements_txt)]


all_ = parse_requirements('requirements.txt')


trainer_extra = []


pytorch_lightning_extra = [
    "pytorch-lightning>=1.5.2",
]

torchserve_extra = [
    "torchserve",
    "torch-model-archiver",
]


setuptools.setup(
    name="padl-extensions",
    version=versions['__version__'],
    author="LF1",
    author_email="contact@lf1.io",
    description="Extensions for pytorch abstractions for deep learning",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/lf1-io/padl-extensions",
    packages=setuptools.find_packages(),
    setup_requires=[],
    license="Apache 2.0",
    classifiers=[
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        'Topic :: Scientific/Engineering',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
    ],
    python_requires='>=3.8',
    install_requires=all_,
    test_suite="tests",
    tests_require=all_ + parse_requirements('requirements-test.txt') + pytorch_lightning_extra
                  + torchserve_extra,
    package_data={'': ['requirements.txt']},
    include_package_data=True,
    extras_require={
        'all': all_
               + pytorch_lightning_extra
               + torchserve_extra
               + trainer_extra,
        'pytorch_lightning': pytorch_lightning_extra,
        'torchserve': torchserve_extra,
        'trainer': trainer_extra,
    }
)
